import numpy as np
import scipy.stats as stats


from popsynth.population_synth import PopulationSynth


class LogNormalPopulation(PopulationSynth):

    def __init__(self, mu, tau, r_max=10, seed=1234, name='_lognorm'):

        PopulationSynth.__init__(self, r_max, seed, name)

        self.set_luminosity_function_parameters(mu=mu, tau=tau)

        self._lf_form = r"\frac{\alpha L_{\rm min}^{\alpha}}{L^{\alpha+1}}"
    

    def phi(self, L):

        return (1./(self.tau * L * np.sqrt(2* np.pi))) * np.exp(-(np.log(L) - self.mu)**2/(2*self.tau**2))

    def draw_luminosity(self, size=1):
    
        return np.random.lognormal(mean=self.mu, sigma=self.tau, size=size)
        
    def __get_mu(self):
             """Calculates the 'mu' property."""
             return self._lf_params['mu']

    def ___get_mu(self):
         """Indirect accessor for 'mu' property."""
         return self.__get_mu()

    def __set_mu(self, mu):
         """Sets the 'mu' property."""
         self.set_luminosity_function_parameters(mu=mu, tau=self.tau)

    def ___set_mu(self, mu):
         """Indirect setter for 'mu' property."""
         self.__set_mu(mu)

    mu = property(___get_mu, ___set_mu,
                     doc="""Gets or sets the mu.""")

    def __get_tau(self):
             """Calculates the 'tau' property."""
             return self._lf_params['tau']

    def ___get_tau(self):
         """Indirect accessor for 'tau' property."""
         return self.__get_tau()

    def __set_tau(self, tau):
         """Sets the 'tau' property."""
         self.set_luminosity_function_parameters(mu=self.mu, tau=tau)

    def ___set_tau(self, tau):
         """Indirect setter for 'tau' property."""
         self.__set_tau(tau)

    tau = property(___get_tau, ___set_tau,
                     doc="""Gets or sets the tau.""")


    



import numpy as np
from popsynth.spherical_population import ConstantSphericalPopulation, ZPowerSphericalPopulation
from popsynth.cosmological_population import SFRPopulation

from popsynth.pareto_population import ParetoPopulation
from popsynth.schechter_population import SchechterPopulation
from popsynth.log10_normal_population import Log10NormalPopulation
from popsynth.log_normal_population import LogNormalPopulation
from popsynth.bpl_population import BPLPopulation

class ParetoHomogeneousSphericalPopulation(ParetoPopulation, ConstantSphericalPopulation):

    def __init__(self, Lambda, Lmin, alpha, r_max=10, seed=1234):

        ParetoPopulation.__init__(
            self, Lmin=Lmin, alpha=alpha, r_max=r_max, seed=seed, name='ParetoHomogeneousSphericalPopulation')
        ConstantSphericalPopulation.__init__(
            self, Lambda=Lambda, r_max=r_max, seed=seed, name='ParetoHomogeneousSphericalPopulation')


class Log10NormalHomogeneousSphericalPopulation(Log10NormalPopulation, ConstantSphericalPopulation):

    def __init__(self, Lambda, mu, tau, r_max=10, seed=1234):

        Log10NormalPopulation.__init__(
            self, mu=mu, tau=tau, r_max=r_max, seed=seed, name='Log10NormalHomogeneousSphericalPopulation')
        ConstantSphericalPopulation.__init__(
            self, Lambda=Lambda, r_max=r_max, seed=seed, name= 'Log10NormalHomogeneousSphericalPopulation')


class Log10NormalZPowerSphericalPopulation(Log10NormalPopulation, ZPowerSphericalPopulation):

    def __init__(self, Lambda, delta, mu, tau, r_max=10, seed=1234):

        Log10NormalPopulation.__init__(
            self, mu=mu, tau=tau, r_max=r_max, seed=seed, name='Log10NormalZPowerSphericalPopulation')
        ZPowerSphericalPopulation.__init__(
            self, Lambda=Lambda, delta=delta, r_max=r_max, seed=seed, name='Log10NormalZPowerSphericalPopulation')

        

class SchechterHomogeneousSphericalPopulation(SchechterPopulation, ConstantSphericalPopulation):

    def __init__(self, Lambda, Lmin, alpha, r_max=10, seed=1234):

        SchechterPopulation.__init__(
            self, Lmin=Lmin, alpha=alpha, r_max=r_max, seed=seed, name='SchechterHomogeneousSphericalPopulation')
        ConstantSphericalPopulation.__init__(
            self, Lambda=Lambda, r_max=r_max, seed=seed, name='SchechterHomogeneousSphericalPopulation')


class ParetoSFRPopulation(ParetoPopulation, SFRPopulation):

    def __init__(self, r0, rise, decay, peak, Lmin, alpha, r_max=10, seed=1234):

        ParetoPopulation.__init__(self, Lmin=Lmin, alpha=alpha, r_max=r_max, seed=seed, name='ParetoSFRPopulation')
        SFRPopulation.__init__(
            self, r0=r0, rise=rise, decay=decay, peak=peak, r_max=r_max, seed=seed, name='ParetoSFRPopulation')


    def generate_stan_code(self, stan_gen, **kwargs):

        ParetoPopulation.generate_stan_code(self, stan_gen, **kwargs)
        SFRPopulation.generate_stan_code(self, stan_gen, **kwargs)
        

class BPLSFRPopulation(BPLPopulation, SFRPopulation):

    def __init__(self, r0, rise, decay, peak, Lmin, alpha, Lbreak, beta, Lmax,  r_max=10, seed=1234):

        BPLPopulation.__init__(self, Lmin=Lmin, alpha=alpha, Lbreak=Lbreak, beta=beta, Lmax=Lmax,  r_max=r_max, seed=seed, name='BPLSFRPopulation')
        SFRPopulation.__init__(
            self, r0=r0, rise=rise, decay=decay, peak=peak, r_max=r_max, seed=seed, name='BPLSFRPopulation')


    def generate_stan_code(self, stan_gen, **kwargs):
        pass
#        ParetoPopulation.generate_stan_code(self, stan_gen, **kwargs)
 #       SFRPopulation.generate_stan_code(self, stan_gen, **kwargs)
        

        
class SchechterSFRPopulation(SchechterPopulation, SFRPopulation):

    def __init__(self, r0, rise, decay, peak, Lmin, alpha, r_max=10, seed=1234):

        SchechterPopulation.__init__(
            self, Lmin=Lmin, alpha=alpha, r_max=r_max, seed=seed, name='SchechterSFRPopulation')
        SFRPopulation.__init__(
            self, r0=r0, rise=rise, decay=decay, peak=peak, r_max=r_max, seed=seed, name='SchechterSFRPopulation')


class LogNormalSFRPopulation(LogNormalPopulation, SFRPopulation):

    def __init__(self, r0, rise, decay, peak, mu, tau, r_max=10, seed=1234):

        LogNormalPopulation.__init__(
            self, mu=mu, tau=tau, r_max=r_max, seed=seed, name='LogNormalSFRPopulation')
        SFRPopulation.__init__(
            self, r0=r0, rise=rise, decay=decay, peak=peak, r_max=r_max, seed=seed, name='LogNormalSFRPopulation')



import numpy as np

def sample_theta_phi(size):
    
    theta = np.arccos(1- 2*np.random.uniform(0.,1., size=size) )
    phi = np.random.uniform(0, 2*np.pi, size=size)
    
    return theta, phi


def xyz(r,theta, phi):
    x = r * np.sin(theta) * np.cos(phi)
    y = r * np.sin(theta) * np.sin(phi)
    z = r * np.cos(theta)
    
    return x,y,z



import numpy as np


from astropy.cosmology import WMAP9 as cosmo

from popsynth.population_synth import PopulationSynth
from popsynth.utils.package_data import copy_package_data

from astropy.constants import c as sol
sol = sol.value


h0 = 69.3
dh = sol * 1.e-3 / h0
Om = 0.286
Om_reduced = ((1-Om)/Om)
Om_sqrt = np.sqrt(Om)
Ode = 1 - Om -(cosmo.Onu0 + cosmo.Ogamma0)


def Phi(x):    
    x2 = x*x
    x3 = x*x*x 
    top = 1. + 1.320*x + 0.441 *x2 + 0.02656*x3
    bottom = 1. + 1.392*x + 0.5121*x2 + 0.03944*x3
    return top/bottom

def xx(z): 
    return Om_reduced / np.power(1+z,3)


def luminosity_distance(z):
    x = xx(z)
    z1 = 1+z
    val = (2 * dh * z1 / Om_sqrt) *( Phi(xx(0)) - 1./(np.sqrt(z1)) * Phi(x)) *3.086E24 # in cm
    return val

def a(z):
    return np.sqrt( np.power(1+z,3.)*Om +Ode)

def comoving_transverse_distance(z):
    return luminosity_distance(z)/(1.+z)

def differential_comoving_volume(z):
    td = (comoving_transverse_distance(z)/3.086E24) 
    return (dh*td*td/a(z)) * 1E-9 # Gpc^3


class CosmologicalPopulation(PopulationSynth):

    def __init__(self, r_max=10, seed=1234, name='cosmo'):

        PopulationSynth.__init__(self, r_max, seed, name)

    def differential_volume(self, z):

        td = (comoving_transverse_distance(z)/3.086E24) 
        return (dh*td*td/a(z)) * 1E-9 # Gpc^3


    def transform(self, L, z):

        return L/(4. * np.pi * luminosity_distance(z)**2)

    def time_adjustment(self, z):
        return (1+z)


    def generate_stan_code(self, stan_gen, **kwargs):

        # add the cosmology code
        copy_package_data('cosmo.stan')
        copy_package_data('cosmo_constants.stan')

        stan_gen.blocks['functions'].add_include('cosmo.stan')
        stan_gen.blocks['transformed_data'].add_include('cosmo_constants.stan')

        if 'distance_flag' in kwargs:
            distance_flag = kwargs['distance_flag']
        else:

            distance_flag = False
            
        if not distance_flag:
            code = """
            // technical varaibales required for the integration
            real x_r[0];
            int x_i[0];

            real zout[1];
            real state0[1];

            // here we precompute things which depend on the data only

            vector[Nz] z2_inverse;
            vector[N] log_p_obs;
            vector[Nz] log_volume_element;
            real total_static_prob;


            zout[1] = z_max;
            state0[1] = 0.0;

            // the detection probability and comoving volume elements
            // for the observed objects is static
            // we precompute and then sum
            log_volume_element = log(differential_comoving_volume(known_z_obs, Om, Ode, hubble_distance, Om_reduced,Om_sqrt,phi_0)) - log(1+z_obs);
            log_p_obs = log_p_det(log_flux_obs, boundary, strength);
            total_static_prob = sum(log_p_obs) + sum(log_volume_element);

            // the transformation of the objects is static
            z2_inverse = transform(to_vector(rep_array(1.,Nn)), known_z_obs , hubble_distance, Om_reduced, Om_sqrt, phi_0);


            """


        else:
            code = """
            // technical varaibales required for the integration
            real x_r[0];
            int x_i[0];

            real zout[1];
            real state0[1];

            // here we precompute things which depend on the data only

            vector[N] z2_inverse;
            vector[N] log_p_obs;
            vector[N] log_volume_element;
            real total_static_prob;


            zout[1] = z_max;
            state0[1] = 0.0;

            // the detection probability and comoving volume elements
            // for the observed objects is static
            // we precompute and then sum
            log_volume_element = log(differential_comoving_volume(z_obs, Om, Ode, hubble_distance, Om_reduced,Om_sqrt,phi_0)) - log(1+z_obs);
            log_p_obs = log_p_det(log_flux_obs, boundary, strength);
            total_static_prob = sum(log_p_obs) + sum(log_volume_element);

            // the transformation of the objects is static
            z2_inverse = transform(to_vector(rep_array(1.,N)), z_obs , hubble_distance, Om_reduced, Om_sqrt, phi_0);


            """

        stan_gen.blocks['transformed_data'].insert_code(code)



        code = """

         // setup for the integral
        real Lambda; // this is total Lambda!
        real params[10];
        real integration_result[1,1];


        vector[M] log_flux_tilde_latent;
        vector[M] log_dndv_tilde;
        vector[M] log_dvdz_tilde;
        vector[M] log_pndet;

        

        """


        stan_gen.blocks['model'].insert_code(code)


        if not distance_flag:

            code = """
        
            log_flux_obs ~ normal(log10(flux_latent), flux_sigma);
            
            // add the differential of the inhomogeneous process on
            // and detection probability
        
            target += total_static_prob;
            target += sum(log(dNdV(known_z_obs, r0, rise, decay, peak)));
            
            
            // unkown z
            target += sum( log( differential_comoving_volume(z, Om, Ode, hubble_distance, Om_reduced,Om_sqrt,phi_0) ) - log(1+z) );
            target += sum( log( dNdV(z, r0, rise, decay, peak) ) );
            """


        else:

            code = """
        
            log_flux_obs ~ normal(log10(flux_latent), flux_sigma);


            // add the differential of the inhomogeneous process on
            // and detection probability

            target += total_static_prob;
            target += sum(log(dNdV(z_obs, r0, rise, decay, peak)));

            """


        stan_gen.blocks['model'].insert_code(code)

        code = """



        log_flux_tilde_latent = log10(transform(lum_tilde_latent, z_tilde, hubble_distance, Om_reduced, Om_sqrt, phi_0));
        

        log_dndv_tilde =  log(dNdV(z_tilde, r0, rise, decay, peak));
        log_dvdz_tilde = log(differential_comoving_volume(z_tilde, Om, Ode, hubble_distance, Om_reduced,Om_sqrt,phi_0))-log(1+z_tilde);
        log_pndet = log_p_ndet(log_flux_tilde, boundary, strength) ;

        for (m in 1:M) {


           target+= log_sum_exp( log_dndv_tilde[m]
                          + log_dvdz_tilde[m]
                          + log_pndet[m]
                          + normal_lpdf(log_flux_tilde[m] | log_flux_tilde_latent[m], flux_sigma)
                          //insert here
                          ,

                          log(Lambda0) + normal_lpdf(log_flux_tilde[m]| log10(boundary),4)+
                          // insert here
                          + uniform_lpdf(z_tilde[m]| 0 ,z_max)

                          );



        }

        // Poisson normalization for the integral rate
        
        // (Distinguishiable) Poisson process model

        params[1] = r0;
        params[2] = rise;
        params[3] = decay;
        params[4] = peak;
        params[5] = Om;
        params[6] = Ode;
        params[7] = hubble_distance;
        params[8] = Om_reduced;
        params[9] = Om_sqrt;
        params[10] = phi_0;


        // integrate the dN/dz to get the normalizing constant for given r0 and alpha
        integration_result = integrate_ode_rk45(N_integrand, state0, 0.0, zout, params, x_r, x_i);
        Lambda = integration_result[1,1];

        
        
        target += - Lambda - Lambda0;
        """

        stan_gen.blocks['model'].insert_code(code)
        
class SFRPopulation(CosmologicalPopulation):

    def __init__(self, r0, rise, decay, peak, r_max=10, seed=1234, name='_sfrcosmo'):

        CosmologicalPopulation.__init__(self, r_max, seed, name)

        self.set_spatial_distribution_params(r0=r0, rise=rise, decay=decay, peak=peak)

        self._spatial_form = r'\rho_0 \frac{1+r \cdot z}{1+ \left(z/p\right)^d}'


    def dNdV(self, z):
        top = 1. + self.rise * z
        bottom = 1+np.power(z/self.peak, self.decay)
    
        return self.r0 * top/bottom

        
    def __get_r0(self):
             """Calculates the 'r0' property."""
             return self._spatial_params['r0']

    def ___get_r0(self):
         """Indirect accessor for 'r0' property."""
         return self.__get_r0()

    def __set_r0(self, r0):
         """Sets the 'r0' property."""
         self.set_spatial_distribution_params(r0=r0, rise=self.rise, decay=self.decay, peak=self.peak)

    def ___set_r0(self, r0):
         """Indirect setter for 'r0' property."""
         self.__set_r0(r0)

    r0 = property(___get_r0, ___set_r0,
                     doc="""Gets or sets the r0.""")

    def __get_rise(self):
             """Calculates the 'rise' property."""
             return self._spatial_params['rise']

    def ___get_rise(self):
         """Indirect accessor for 'rise' property."""
         return self.__get_rise()

    def __set_rise(self, rise):
         """Sets the 'rise' property."""
         self.set_spatial_distribution_params(r0=self.r0, rise=rise, decay=self.decay, peak=self.peak)

    def ___set_rise(self, rise):
         """Indirect setter for 'rise' property."""
         self.__set_rise(rise)

    rise = property(___get_rise, ___set_rise,
                     doc="""Gets or sets the rise.""")

    
    def __get_decay(self):
             """Calculates the 'decay' property."""
             return self._spatial_params['decay']

    def ___get_decay(self):
         """Indirect accessor for 'decay' property."""
         return self.__get_decay()

    def __set_decay(self, decay):
         """Sets the 'decay' property."""
         self.set_spatial_distribution_params(r0=self.r0, rise=self.rise, decay=decay, peak=self.peak)

    def ___set_decay(self, decay):
         """Indirect setter for 'decay' property."""
         self.__set_decay(decay)

    decay = property(___get_decay, ___set_decay,
                     doc="""Gets or sets the decay.""")


    def __get_peak(self):
        """Calculates the 'peak' property."""
        return self._spatial_params['peak']

    def ___get_peak(self):
         """Indirect accessor for 'peak' property."""
         return self.__get_peak()

    def __set_peak(self, peak):
         """Sets the 'peak' property."""
         self.set_spatial_distribution_params(r0=self.r0, rise=self.rise, decay=self.decay, peak=peak)

    def ___set_peak(self, peak):
         """Indirect setter for 'peak' property."""
         self.__set_peak(peak)

    peak = property(___get_peak, ___set_peak,
                     doc="""Gets or sets the peak.""")


    def generate_stan_code(self, stan_gen, **kwargs):


        CosmologicalPopulation.generate_stan_code(self, stan_gen, **kwargs)

        copy_package_data('sfr_functions.stan')
        stan_gen.blocks['functions'].add_include('sfr_functions.stan')

        



import numpy as np

import abc

class AuxiliarySampler(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, name, sigma):

        self._sigma = sigma
        
        self._name = name
        self._obs_name = '%s_obs' %name

        self._obs_values = None
        self._true_values = None
        

    def set_luminosity(self, luminosity):

        self._luminosity = luminosity

    def set_distance(self, distance):

        self._distance = distance
        
    @property
    def name(self):
        return self._name

    @property
    def obs_name(self):
        return self._obs_name

    @property
    def sigma(self):
        return self._sigma
    
    @property
    def true_values(self):
        return self._true_values

    @property
    def obs_values(self):
        return self._obs_values
    
    @property
    
    @abc.abstractmethod
    def true_sampler(self, size=1):

        pass

    @abc.abstractmethod
    def observation_sampler(self, size=1):

        pass

    
class DerivedLumAuxSampler(AuxiliarySampler):

    def __init__(self, name, sigma):

        super(DerivedLumAuxSampler, self).__init__(name, sigma)

    def compute_luminosity(self):
        pass



import numpy as np

from popsynth.population_synth import PopulationSynth


class ParetoPopulation(PopulationSynth):

    def __init__(self, Lmin, alpha, r_max=10, seed=1234, name='_pareto'):

        PopulationSynth.__init__(self, r_max, seed, name)

        self.set_luminosity_function_parameters(Lmin=Lmin, alpha=alpha)




        self._lf_form = r"\frac{\alpha L_{\rm min}^{\alpha}}{L^{\alpha+1}}"
    

    def phi(self, L):

        out = np.zeros_like(L)

        idx = L>=self.Lmin
        
        out[idx] =  self.alpha*self.Lmin**self.alpha / L[idx]**(self.alpha+1)

        return out
        
    def draw_luminosity(self, size=1):
    
        return (np.random.pareto(self._lf_params['alpha'], size) + 1) * self._lf_params['Lmin']
        
    def __get_Lmin(self):
             """Calculates the 'Lmin' property."""
             return self._lf_params['Lmin']

    def ___get_Lmin(self):
         """Indirect accessor for 'Lmin' property."""
         return self.__get_Lmin()

    def __set_Lmin(self, Lmin):
         """Sets the 'Lmin' property."""
         self.set_luminosity_function_parameters(alpha=self.alpha, Lmin=Lmin)

    def ___set_Lmin(self, Lmin):
         """Indirect setter for 'Lmin' property."""
         self.__set_Lmin(Lmin)

    Lmin = property(___get_Lmin, ___set_Lmin,
                     doc="""Gets or sets the Lmin.""")


    
    def __get_alpha(self):
             """Calculates the 'alpha' property."""
             return self._lf_params['alpha']

    def ___get_alpha(self):
         """Indirect accessor for 'alpha' property."""
         return self.__get_alpha()

    def __set_alpha(self, alpha):
         """Sets the 'alpha' property."""
         self.set_luminosity_function_parameters(alpha=alpha,Lmin=self.Lmin)

    def ___set_alpha(self, alpha):
         """Indirect setter for 'alpha' property."""
         self.__set_alpha(alpha)

    alpha = property(___get_alpha, ___set_alpha,
                     doc="""Gets or sets the alpha.""")



    def generate_stan_code(self, stan_gen, **kwargs):

        pass




import pkg_resources
import os
from shutil import copyfile

def get_path_of_data_file(data_file):
    file_path = pkg_resources.resource_filename("popsynth", 'data/%s' % data_file)

    return file_path


def get_path_of_data_dir():
    file_path = pkg_resources.resource_filename("popsynth", 'data')

    return file_path

def copy_package_data(data_file):

    data_file_path = get_path_of_data_file(data_file)
    copyfile(data_file_path, './%s'%data_file)
    

    



# def get_path_of_user_dir():
#     """
#     Returns the path of the directory containing the user data (~/.popsynth)

#     :return: an absolute path
#     """

#     return os.path.abspath(os.path.expanduser('~/.popsynth'))



import numpy as np

from popsynth.population_synth import PopulationSynth


class SphericalPopulation(PopulationSynth):

    def __init__(self, r_max=10, seed=1234, name='_sphere'):

        PopulationSynth.__init__( self, r_max, seed, name)

    def differential_volume(self, r):

        return 4 * np.pi * r * r

    def transform(self, L, r):

        return L/(4. * np.pi * (r+1) * (r+1))
    
class ConstantSphericalPopulation(SphericalPopulation):


    def __init__(self, Lambda=1. ,r_max = 10., seed=1234, name='_cons_sphere'):

        super(ConstantSphericalPopulation, self).__init__(r_max, seed, name)


        self.set_spatial_distribution_params(Lambda=Lambda)


        self._spatial_form = r'\Lambda'
        
    def __get_Lambda(self):
             """Calculates the 'Lambda' property."""
             return self._spatial_params['Lambda']

    def ___get_Lambda(self):
         """Indirect accessor for 'Lambda' property."""
         return self.__get_Lambda()

    def __set_Lambda(self, Lambda):
         """Sets the 'Lambda' property."""
         self.set_spatial_distribution_params(Lambda=Lambda)

    def ___set_Lambda(self, Lambda):
         """Indirect setter for 'Lambda' property."""
         self.__set_Lambda(Lambda)

    Lambda = property(___get_Lambda, ___set_Lambda,
                     doc="""Gets or sets the Lambda.""")
        
    
    def dNdV(self, distance):

        return self._spatial_params['Lambda']

class ZPowerSphericalPopulation(ConstantSphericalPopulation):


    def __init__(self, Lambda=1., delta=1. ,r_max = 10., seed=1234, name='_cons_sphere'):

        self.set_spatial_distribution_params(Lambda=Lambda, delta=delta)

        super(ZPowerSphericalPopulation, self).__init__(Lambda, r_max ,seed, name)


        


        self._spatial_form = r'\Lambda (z+1)^{\delta}'
        
    def __get_delta(self):
             """Calculates the 'delta' property."""
             return self._spatial_params['delta']

    def ___get_delta(self):
         """Indirect accessor for 'delta' property."""
         return self.__get_delta()

    def __set_delta(self, delta):
         """Sets the 'delta' property."""
         self.set_spatial_distribution_params(delta=delta)

    def ___set_delta(self, delta):
         """Indirect setter for 'delta' property."""
         self.__set_delta(delta)

    delta = property(___get_delta, ___set_delta,
                     doc="""Gets or sets the delta.""")
        
    
    def dNdV(self, distance):

        return self._spatial_params['Lambda'] * np.power(distance+1., self._spatial_params['delta'])





from popsynth.population_synth import PopulationSynth
from popsynth.spherical_population import ConstantSphericalPopulation, SphericalPopulation
from popsynth.cosmological_population import CosmologicalPopulation, SFRPopulation
from popsynth.pareto_population import ParetoPopulation
from popsynth.schechter_population import SchechterPopulation

from popsynth.population import Population
from popsynth.auxiliary_sampler import AuxiliarySampler

from popsynth import synths
#from popsynth.synths import ParetoConstantSphericalPopulation, ParetoSFRPopulation, SchechterSFRPopulation





import numpy as np

chance = np.random.uniform(0,1,size=1)

if chance <= 0.8:

    from IPython.display import display, YouTubeVideo


    
    display(YouTubeVideo('NFTaiWInZ44'))



import h5py
import importlib
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import importlib
importlib.import_module('mpl_toolkits.mplot3d').Axes3D
import pandas as pd
from IPython.display import display, Math, Markdown

from popsynth.utils.spherical_geometry import sample_theta_phi, xyz
from stan_utility import StanGenerator


green = '#24B756'
green_highlight = '#07A23B'
orange = '#FA7031'
orange_highlight = '#DD4B09'
blue = '#1E9A91'
blue_highlight = '#06887F'


class Population(object):

    def __init__(self,
                 luminosities,
                 distances,
                 known_distances,
                 known_distance_idx,
                 unknown_distance_idx,
                 fluxes,
                 flux_obs,
                 selection,
                 flux_sigma,
                 r_max,
                 boundary,
                 strength,
                 n_model,
                 lf_params,
                 spatial_params=None,
                 model_spaces=None,
                 seed=1234,
                 name=None,
                 spatial_form=None,
                 lf_form=None,
                 auxiliary_quantities=None):

        """
        A Population of detected objects
        """

        
        self._luminosities = luminosities


        self._distances = distances
        self._known_distances = known_distances
        self._known_distance_idx =known_distance_idx
        self._unknown_distance_idx = unknown_distance_idx

        assert len(known_distances) + len(unknown_distance_idx) == sum(selection), 'the distances are not the correct size'
        
        
        self._fluxes = fluxes
        self._flux_obs = flux_obs
        self._selection = selection
        self._flux_sigma = flux_sigma
        
        self._r_max = r_max
        
        self._boundary = boundary
        self._strength = strength
        self._seed = seed
        self._n_model = n_model
        self._name = name
        self._spatial_form = spatial_form
        self._lf_form = lf_form

        self._flux_selected = flux_obs[selection]
        self._distance_selected = distances[selection]
        self._luminosity_selected = luminosities[selection]

        self._flux_hidden = flux_obs[~selection]
        self._distance_hidden = distances[~selection]
        self._luminosity_hidden = luminosities[~selection]

        self._lf_params = lf_params
        self._spatial_params = spatial_params

        self._model_spaces = model_spaces

        if auxiliary_quantities is not None:

            for k, v in auxiliary_quantities.items():

                setattr(self, k, v['true_values'])
                setattr(self, '%s_obs' % k, v['obs_values'])
                setattr(self, '%s_selected' % k, v['obs_values'][selection])

        self._auxiliary_quantites = auxiliary_quantities

        if model_spaces is not None:

            for k, v in model_spaces.items():

                assert len(v) == n_model

    @property
    def luminosities(self):
        return self._luminosities

    @property
    def fluxes(self):
        return self._fluxes

    @property
    def distances(self):
        return self._distances

    @property
    def known_distances(self):
        return self._known_distances
    
    @property
    def selection(self):
        return self._selection

    @property
    def flux_observed_all(self):
        return self._flux_obs

    @property
    def selected_fluxes(self):
        return self._flux_selected

    @property
    def selected_distances(self):
        return self._distance_selected

    @property
    def luminosity_parameters(self):
        return self._lf_params

    @property
    def spatial_parameters(self):
        return self._spatial_params


    def _prob_det(self, x, boundary, strength):
        """
        Soft detection threshold

        :param x: values to test
        :param boundary: mean value of the boundary
        :param strength: the strength of the threshold
        """


        return sf.expit(strength * (x - boundary))

    def recompute_selection(self, boundary, stength):

        raise NotImplementedError('not ready for this yet')
        
        # compute the detection probability  for the observed values
        detection_probability = self._prob_det(log10_fluxes_obs, np.log10(boundary), strength)

        # now select them
        selection = []
        for p in detection_probability:

            # make a bernoulli draw given the detection probability
            if stats.bernoulli.rvs(p) == 1:

                selection.append(True)

            else:

                selection.append(False)

        selection = np.array(selection)

        if sum(selection) == n:
            print('NO HIDDEN OBJECTS')
            
        
        print('Deteced %d objects or to a distance of %.2f' %(sum(selection), max(distances[selection])))


    def generate_stan_code(self, model_name='model', population_synth=None):
        

        stan_gen = StanGenerator(model_name, data_size='N')

        # First we add on the population level parameters

        # luminosity function
        for k,v in self._lf_params.items():

            stan_gen.add_parameters(k)

                # luminosity function
        for k,v in self._spatial_params.items():

            stan_gen.add_parameters(k)


        
        # now the basic data

        # assume homoskedastic flux_sigma
        stan_gen.add_data('flux_sigma','z_max', 'boundary', 'strength')

        stan_gen.add_data('M', stan_type='int')
        # add vector data
        stan_gen.add_standard_vector_data('log_flux_obs')

        # add typical params... I assume mixtures from now on
        stan_gen.add_standard_vector_parameters('luminosity_latent', lower_bound='0')

        
        
        # mixture stuff
        stan_gen.add_parameters('Lambda0', lower_bound='0')

        stan_gen.add_vector_parameters('luminosity_tilde_latent',size='M', lower_bound=0)
        stan_gen.add_vector_parameters('log_flux_tilde',size='M', lower_bound=0)

        
        distance_flag = False
        if len(self._known_distances) == len(self._distance_selected):
            # ok, we know all the distances so things will be normal
            stan_gen.add_standard_vector_data('z_obs')
            distance_flag = True

        else:

            # now we needed to add the unknown distance stuff
            stan_gen.add_data( 'Nz','Nnz',stan_type='int')
            
            stan_gen.add_vector_data('z_idx',stan_type='int',size='Nz')
            stan_gen.add_vector_data('z_nidx',stan_type='int',size='Nnz')
            stan_gen.add_vector_data('known_z_obs',size='Nz')

            stan_gen.add_vector_parameters('z',size='Nz',lower_bound='0',upper_bound='z_max')
            
        # now we deal with the aux

        stan_gen.add_data('N_model',stan_type='int')
        for k, v in self._model_spaces.items():

            stan_gen.add_vector_data(k,size='N_model')
            
        
        for k, v in self._auxiliary_quantites.items():

            stan_gen.add_standard_vector_data('%s_obs' % k)
            stan_gen.add_data('%s_sigma' % k)
            stan_gen.add_standard_vector_parameters('%s_latent'%k , lower_bound=0)
            # mixture stuff
            stan_gen.add_vector_parameters('%s_tilde_latent'%k , size='M', lower_bound=0)
            stan_gen.add_vector_parameters('%s_tilde_obs'%k,size='M', lower_bound=0)
        

        if population_synth is None:
            print('Will not generate population code')

        else:


            population_synth.generate_stan_code(stan_gen=stan_gen, distance_flag=distance_flag)

        stan_gen.write_stan_code()

        return stan_gen
        

    
    def to_stan_data(self):
        """
        Create Stan input
        """

        # create a dict for Stan
        output = dict(
            N=self._selection.sum(),
            Nz=len(self._known_distances),
            Nnz = len(self._unknown_distance_idx),
            z_obs=self._distance_selected,
            known_z_obs = self._known_distances,
            z_idx = self._known_distance_idx+1, # stan indexing
            z_nidx = self._unknown_distance_idx+1, #stan indexing
            log_flux_obs=np.log10(self._flux_selected),
            flux_sigma=self._flux_sigma,
            z_max=self._r_max,
            N_model=self._n_model,
            boundary=self._boundary,
            strength=self._strength)

        # now append the model spaces
        for k, v in self._model_spaces.items():

            output[k] = v

        for k, v in self._auxiliary_quantites.items():

            output['%s_obs' % k] = v['obs_values'][self._selection]
            output['%s_sigma' % k] = v['sigma']

        return output

    def writeto(self, file_name):

        with h5py.File(file_name, 'w') as f:

            spatial_grp = f.create_group('spatial_params')

            for k, v in self._spatial_params.items():

                spatial_grp.create_dataset(k, data=np.array([v]), compression='lzf')

            lum_grp = f.create_group('lf_params')

            for k, v in self._lf_params.items():

                lum_grp.create_dataset(k, data=np.array([v]), compression='lzf')

            f.attrs['name'] = np.string_(self._name)
            f.attrs['lf_form'] = np.string_(self._lf_form)
            f.attrs['spatial_form'] = np.string_(self._spatial_form)
            f.attrs['flux_sigma'] = self._flux_sigma
            f.attrs['n_model'] = self._n_model
            f.attrs['r_max'] = self._r_max
            f.attrs['boundary'] = self._boundary
            f.attrs['strength'] = self._strength
            f.attrs['seed'] = int(self._seed)

            f.create_dataset('luminosities', data=self._luminosities, compression='lzf')
            f.create_dataset('distances', data=self._distances, compression='lzf')
            f.create_dataset('known_distances', data=self._known_distances, compression='lzf')
            f.create_dataset('known_distance_idx', data=self._known_distance_idx, compression='lzf')
            f.create_dataset('unknown_distance_idx', data=self._unknown_distance_idx, compression='lzf')
            f.create_dataset('fluxes', data=self._fluxes, compression='lzf')
            f.create_dataset('flux_obs', data=self._flux_obs, compression='lzf')
            f.create_dataset('selection', data=self._selection, compression='lzf')

            aux_grp = f.create_group('auxiliary_quantities')

            for k, v in self._auxiliary_quantites.items():

                q_grp = aux_grp.create_group(k)
                q_grp.create_dataset('true_values', data=v['true_values'], compression='lzf')
                q_grp.create_dataset('obs_values', data=v['obs_values'], compression='lzf')

                q_grp.attrs['sigma'] = v['sigma']

            model_grp = f.create_group('model_spaces')

            for k, v in self._model_spaces.items():

                model_grp.create_dataset(k, data=v, compression='lzf')

    @classmethod
    def from_file(cls, file_name):

        with h5py.File(file_name, 'r') as f:

            spatial_params = {}
            lf_params = {}

            for key in f['spatial_params'].keys():

                spatial_params[key] = f['spatial_params'][key].value[0]

            for key in f['lf_params'].keys():

                lf_params[key] = f['lf_params'][key].value[0]

            flux_sigma = f.attrs['flux_sigma']
            boundary = f.attrs['boundary']
            strength = f.attrs['strength']
            n_model = f.attrs['n_model']
            r_max = f.attrs['r_max']
            seed = int(f.attrs['seed'])
            name = f.attrs['name']
            lf_form = str(f.attrs['lf_form'])
            spatial_form = str(f.attrs['spatial_form'])

            luminosities = f['luminosities'].value
            distances = f['distances'].value

            # right now this is just for older pops
            try:
                known_distances = f['known_distances'].value
                known_distance_idx = ( f['known_distance_idx'].value).astype(int)
                unknown_distance_idx = ( f['unknown_distance_idx'].value).astype(int)

            except:

                known_distances = None
                known_distance_idx = None
                unknown_distance_idx = None
                
            fluxes = f['fluxes'].value
            flux_obs = f['flux_obs'].value
            selection = f['selection'].value

            model_spaces = {}

            for k in f['model_spaces'].keys():

                model_spaces[str(k)] = f['model_spaces'][k].value

            auxiliary_quantities = {}

            for k in f['auxiliary_quantities'].keys():

                auxiliary_quantities[str(k)] = {
                    'true_values': f['auxiliary_quantities'][k]['true_values'].value,
                    'obs_values': f['auxiliary_quantities'][k]['obs_values'].value,
                    'sigma': f['auxiliary_quantities'][k].attrs['sigma']
                }

        return cls(
            luminosities=luminosities,
            distances=distances,
            known_distances=known_distances,
            known_distance_idx=known_distance_idx,
            unknown_distance_idx=unknown_distance_idx,
            fluxes=fluxes,
            flux_obs=flux_obs,
            selection=selection,
            flux_sigma=flux_sigma,
            n_model=n_model,
            r_max=r_max,
            boundary=boundary,
            strength=strength,
            lf_params=lf_params,
            spatial_params=spatial_params,
            model_spaces=model_spaces,
            seed=seed,
            name=name,
            spatial_form=spatial_form,
            lf_form=lf_form,
            auxiliary_quantities=auxiliary_quantities)

    def display(self):
        """
        Display the simulation parameters
        
        """

        info = '### %s simulation\nDetected %d out of %d objects' % (self._name, sum(self._selection), len(
            self._fluxes))

        display(Markdown(info))

        out = {'parameter': [], 'value': []}

        display(Markdown('## Luminosity Function'))
        for k, v in self._lf_params.items():

            out['parameter'].append(k)
            out['value'].append(v)

        display(Math(self._lf_form))
        display(pd.DataFrame(out))
        out = {'parameter': [], 'value': []}

        display(Markdown('## Spatial Function'))

        for k, v in self._spatial_params.items():

            out['parameter'].append(k)
            out['value'].append(v)

        display(Math(self._spatial_form))
        display(pd.DataFrame(out))

    def display_true_fluxes(self, ax=None, flux_color=orange, **kwargs):

        if ax is None:
            fig, ax = plt.subplots()

        else:

            fig = ax.get_figure()

        ax.scatter(self._distances, self._fluxes, alpha=.5, color=flux_color, edgecolors='none',label='True flux', **kwargs)

        ax.axhline(self._boundary, color='grey', zorder=-5000, ls='--')

        #ax.set_xscale('log')
        ax.set_yscale('log')

        ax.set_ylim(bottom=min([self._fluxes.min(), self._flux_selected.min()]))

        ax.set_xlabel('distance')
        ax.set_ylabel('flux')

    def display_obs_fluxes(self, ax=None, flux_color=green, **kwargs):

        
        if ax is None:
            fig, ax = plt.subplots()

        else:

            fig = ax.get_figure()

        ax.scatter(self._distance_selected, self._flux_selected, alpha=.8,
                   color=flux_color, edgecolors='none',label='Detected flux', **kwargs)

        ax.axhline(self._boundary, color='grey', zorder=-5000, ls='--')
        #ax.set_xscale('log')
        ax.set_yscale('log')

        ax.set_ylim(bottom=min([self._fluxes.min(), self._flux_selected.min()]))
        ax.set_xlim(right=self._r_max)

        ax.set_xlabel('distance')
        ax.set_ylabel('flux')
        return fig
        
    def display_fluxes(self, ax=None, true_color=orange, obs_color=green,arrow_color='k',with_arrows=True, **kwargs):

        if ax is None:
            fig, ax = plt.subplots()

        else:

            fig = ax.get_figure()

        self.display_true_fluxes(ax=ax, flux_color=true_color, **kwargs)
        self.display_obs_fluxes(ax=ax, flux_color=obs_color, **kwargs)

        if with_arrows:
            for start, stop, z in zip(self._fluxes[self._selection], self._flux_selected, self._distance_selected):

                x = z
                y = start
                dx = 0
                dy = stop - start

                ax.arrow(x, y, dx, dy, color=arrow_color, head_width=0.05, head_length=0.2 * np.abs(dy), length_includes_head=True)

        return fig
            
    def display_obs_fluxes_sphere(self, ax=None, cmap='magma', distance_transform=None, use_log=False, **kwargs):

        if ax is None:
            fig, ax = plt.subplots(subplot_kw=dict(projection='3d'))

        else:

            fig = ax.get_figure()

        n = sum(self._selection)

        theta, phi = sample_theta_phi(n)

        if distance_transform is not None:

            distance = distance_transform(self._distance_selected)

        else:

            distance = self._distance_selected

        x, y, z = xyz(distance, theta, phi)

        R = self._r_max

        u = np.linspace(0, 2 * np.pi, 100)
        v = np.linspace(0, np.pi, 100)
        x2 = R * np.outer(np.cos(u), np.sin(v))
        y2 = R * np.outer(np.sin(u), np.sin(v))
        z2 = R * np.outer(np.ones(np.size(u)), np.cos(v))

        if use_log:

            x = np.log10(x)
            y = np.log10(y)
            z = np.log10(z)

            x2 = np.log10(x2)
            y2 = np.log10(y2)
            z2 = np.log10(z2)

            R = np.log10(R)

        ax.scatter3D(
            x,
            y,
            z,
            c=self._flux_selected,
            cmap=cmap,
            norm=mpl.colors.LogNorm(vmin=min(self._flux_selected), vmax=max(self._flux_selected)),
            **kwargs)

        ax.plot_wireframe(x2, y2, z2, color='grey', alpha=0.9, rcount=4, ccount=2)

        ax._axis3don = False
        ax.set_xlim(-R, R)
        ax.set_ylim(-R, R)
        ax.set_zlim(-R, R)

    def display_fluxes_sphere(self, ax=None, cmap='magma', distance_transform=None, use_log=False, **kwargs):

        if ax is None:
            fig, ax = plt.subplots(subplot_kw=dict(projection='3d'))

        else:

            fig = ax.get_figure()

        n = len(self._fluxes)

        theta, phi = sample_theta_phi(n)

        if distance_transform is not None:

            distance = distance_transform(self._distances)

        else:

            distance = self._distances

        x, y, z = xyz(distance, theta, phi)

        R = self._r_max

        u = np.linspace(0, 2 * np.pi, 100)
        v = np.linspace(0, np.pi, 100)
        x2 = R * np.outer(np.cos(u), np.sin(v))
        y2 = R * np.outer(np.sin(u), np.sin(v))
        z2 = R * np.outer(np.ones(np.size(u)), np.cos(v))

        if use_log:

            x = np.log10(x)
            y = np.log10(y)
            z = np.log10(z)

            x2 = np.log10(x2)
            y2 = np.log10(y2)
            z2 = np.log10(z2)

            R = np.log10(R)

        ax.scatter3D(
            x,
            y,
            z,
            c=self._flux_obs,
            cmap=cmap,
            norm=mpl.colors.LogNorm(vmin=min(self._fluxes), vmax=max(self._fluxes)), **kwargs)

        ax.plot_wireframe(x2, y2, z2, color='grey', alpha=0.9, rcount=4, ccount=2)

        ax._axis3don = False
        ax.set_xlim(-R, R)
        ax.set_ylim(-R, R)
        ax.set_zlim(-R, R)

        return fig

    def display_hidden_fluxes_sphere(self, ax=None, cmap='magma', distance_transform=None, use_log=False, **kwargs):

        # do not display if there is nothing hidden
        if len(self._selection) == sum(self._selection):

            return 

        
        
        if ax is None:
            fig, ax = plt.subplots(subplot_kw=dict(projection='3d'))

        else:

            fig = ax.get_figure()

        n = len(self._flux_hidden)

        theta, phi = sample_theta_phi(n)

        if distance_transform is not None:

            distance = distance_transform(self._distance_hidden)

        else:

            distance = self._distance_hidden

        x, y, z = xyz(distance, theta, phi)

        R = self._r_max

        u = np.linspace(0, 2 * np.pi, 100)
        v = np.linspace(0, np.pi, 100)
        x2 = R * np.outer(np.cos(u), np.sin(v))
        y2 = R * np.outer(np.sin(u), np.sin(v))
        z2 = R * np.outer(np.ones(np.size(u)), np.cos(v))

        if use_log:

            x = np.log10(x)
            y = np.log10(y)
            z = np.log10(z)

            x2 = np.log10(x2)
            y2 = np.log10(y2)
            z2 = np.log10(z2)
            R = np.log10(R)

        ax.scatter3D(
            x,
            y,
            z,
            c=self._flux_hidden,
            cmap=cmap,
            norm=mpl.colors.LogNorm(vmin=min(self._flux_hidden), vmax=max(self._flux_hidden)),
            **kwargs
            
        )

        ax.plot_wireframe(x2, y2, z2, color='grey', alpha=0.9, rcount=4, ccount=2)

        ax._axis3don = False
        # ax.set_xlim(-R, R)
        # ax.set_ylim(-R, R)
        # ax.set_zlim(-R, R)

    def display_flux_sphere(self,
                            ax=None,
                            seen_cmap='magma',
                            unseen_cmap='Greys',
                            distance_transform=None,
                            use_log=False,
                            **kwargs
    ):

        if ax is None:
            fig, ax = plt.subplots(subplot_kw=dict(projection='3d'))

        else:
            fig = ax.get_figure()

        self.display_obs_fluxes_sphere(ax=ax, cmap=seen_cmap,
                                       distance_transform=distance_transform,
                                       use_log=use_log,
                                       **kwargs)
        
        self.display_hidden_fluxes_sphere(ax=ax, cmap=unseen_cmap,
                                          distance_transform=distance_transform,
                                          use_log=use_log,
                                          **kwargs)




        return fig
    def display_luminosty(self, ax=None, **kwargs):

        if ax is None:
            fig, ax = plt.subplots()

        else:

            fig = ax.get_figure()
            
        bins = np.logspace(np.log10(self._luminosities.min()), np.log10(self._luminosities.max()), 30)

        ax.hist(self._luminosities, bins=bins, normed=True, facecolor=orange, edgecolor=orange_highlight, lw=1.5)

        ax.set_xscale('log')
        ax.set_yscale('log')
        ax.set_xlabel('L')

    def display_distances(self, ax=None):

        if ax is None:
            fig, ax = plt.subplots()

        else:
            fig = ax.get_figure()

        bins = np.linspace(0, self._r_max, 40)

        ax.hist(self._distances, bins=bins, facecolor=green, edgecolor=green_highlight, lw=1.5, label='Total Pop.')
        ax.hist(
            self._distance_selected,
            bins=bins,
            facecolor=blue,
            edgecolor=blue_highlight,
            lw=1.5,
            alpha=1,
            label='Obs. Pop.')

        ax.set_xlabel('z')
        ax.legend()
        #sns.despine(offset=5, trim=True);



# Scientific libraries
import numpy as np
import scipy.stats as stats
import scipy.special as sf
import scipy.integrate as integrate

import abc
from IPython.display import display, Math, Markdown

import h5py
import pandas as pd

from astropy.cosmology import WMAP9 as cosmo

from astropy.constants import c as sol
sol = sol.value

from popsynth.population import Population
from popsynth.auxiliary_sampler import DerivedLumAuxSampler
from popsynth.utils.progress_bar import progress_bar




class PopulationSynth(object):
    __metaclass__ = abc.ABCMeta
    
    def __init__(self, r_max=10, seed = 1234, name='no_name'):
        """
        
        """
        self._n_model = 500
        self._seed = int(seed)
        self._model_spaces = {}
        self._auxiliary_observations = {}
        self._name = name

        self._r_max = r_max

        self._has_derived_luminosity = False
        self._derived_luminosity_sampler = None

    def set_luminosity_function_parameters(self, **lf_params):
        """
        Set the luminosity function parameters as keywords
        """

        try:
            for k,v in lf_params.items():

                if k in self._lf_params:
                    self._lf_params[k] = v

                else:
                    RuntimeWarning('%s was not originally in the parameters... ignoring.'%k)
                    
        except:

            # we have not set params before
            
            self._lf_params = lf_params
        

    def set_spatial_distribution_params(self, **spatial_params):
        """
        Set the spatial parameters as keywords
        """

        try:

            for k,v in spatial_params.items():

                if k in self._spatial_params:
                    self._spatial_params[k] = v
                else:
                    RuntimeWarning('%s was not originally in the parameters... ignoring.'%k)
                    
                    
        except:

            # we have not set these before
            
            self._spatial_params = spatial_params

    def add_model_space(self, name, start, stop, log=True):
        """
        Add a model space for stan generated quantities
    
        :param name: name that Stan will use
        :param start: start of the grid
        :param stop: stop of the grid
        :param log: use log10 or not

        """
        if log:
            space = np.logspace(np.log10(start), np.log10(stop), self._n_model)
 
        else:

            space = np.linspace(start, stop, self._n_model)

        self._model_spaces[name] = space

    def add_observed_quantity(self, auxiliary_sampler):

        if isinstance(auxiliary_sampler, DerivedLumAuxSampler):
            self._has_derived_luminosity = True
            self._derived_luminosity_sampler = auxiliary_sampler

        else:
            

            self._auxiliary_observations[auxiliary_sampler.name] = auxiliary_sampler
        
        
        
    @property
    def name(self):
        return self._name

    # The following methods must be implemented in subclasses
    
    @abc.abstractmethod
    def phi(self, L):
        pass
        
    @abc.abstractmethod
    def differential_volume(self, distance):
        pass

    @abc.abstractmethod
    def dNdV(self, distance):

        pass

    def time_adjustment(self, r):

        return 1.

    def draw_distance(self, size):
        """
        Draw the distances from the specified dN/dr model
        """

        # create a callback for the sampler
        dNdr = lambda r: self.dNdV(r) * self.differential_volume(r) / self.time_adjustment(r)

        # find the maximum point
        tmp = np.linspace(0, self._r_max, 1E5)
        ymax = np.max(dNdr(tmp))

        # rejection sampling the distribution
        r_out = []
        with progress_bar(size, title='Drawing distances') as pbar:
            for i in range(size):
                flag = True
                while flag:

                    # get am rvs from 0 to the max of the function
                    
                    y = np.random.uniform(low=0, high=ymax)

                    # get an rvs from 0 to the maximum distance
                    
                    r = np.random.uniform(low=0, high=self._r_max)

                    # compare them
                    
                    if y < dNdr(r):
                        r_out.append(r)
                        flag = False
                pbar.increase()
                
        return np.array(r_out)

    @abc.abstractmethod
    def draw_luminosity(self, size):
        pass

    @abc.abstractmethod
    def transform(self, flux, distance):
        pass

    def _prob_det(self, x, boundary, strength):
        """
        Soft detection threshold

        :param x: values to test
        :param boundary: mean value of the boundary
        :param strength: the strength of the threshold
        """


        return sf.expit(strength * (x - boundary))

    def draw_log10_fobs(self, f, f_sigma, size=1):
        """
        draw the log10 of the the fluxes
        """

        log10_f = np.log10(f)

        # sample from the log distribution to keep positive fluxes
        log10_fobs = log10_f + np.random.normal(loc=0, scale=f_sigma, size=size)

        return log10_fobs

    def draw_survey(self, boundary, flux_sigma=1., strength=10., hard_cut=False, distance_probability=None):
        """
        Draw the total survey and return a Population object

        :param boundary: the mean boundary for flux selection
        :param flux_sigma: the homoskedastic sigma for the flux in log10 space
        :param strength: the log10 strength of the inv logit selection
        :return: a Population object
        """

        # set the random seed
        
        np.random.seed(self._seed)

        # create a callback of the integrand 
        dNdr = lambda r: self.dNdV(r) * self.differential_volume(r) / self.time_adjustment(r)

        # integrate the population to determine the true number of
        # objects
        N = integrate.quad(dNdr, 0., self._r_max)[0]
        
        # this should be poisson distributed
        n = np.random.poisson(N)
        distances = self.draw_distance(size=n)
        
        print('Expecting %d total objects'%n)


        # first check if the auxilliary samplers
        # compute the luminosities
        auxiliary_quantities = {}
        if self._has_derived_luminosity:

            print('Sampling %s' % self._derived_luminosity_sampler.name )
            self._derived_luminosity_sampler.set_distance(distances)

            # sample the true and obs
            # values which are held internally
            self._derived_luminosity_sampler.true_sampler(size=n)
            self._derived_luminosity_sampler.observation_sampler(size=n)

            # check to make sure we sampled!
            assert self._derived_luminosity_sampler.true_values is not None and len(self._derived_luminosity_sampler.true_values) == n
            assert self._derived_luminosity_sampler.obs_values is not None and len(self._derived_luminosity_sampler.obs_values) == n

            # append these values to a dict
            auxiliary_quantities[self._derived_luminosity_sampler.name] = {'true_values': self._derived_luminosity_sampler.true_values,
                                       'obs_values': self._derived_luminosity_sampler.obs_values,
                                       'sigma': self._derived_luminosity_sampler.sigma }

            print('Getting luminosity from derived sampler')
            luminosities = self._derived_luminosity_sampler.compute_luminosity()
            
        else:
            # draw all the values
            luminosities = self.draw_luminosity(size=n)
 

        # transform the fluxes
        fluxes = self.transform(luminosities, distances)


        # now sample any auxilary quantities
        # if needed
        
        

        for k,v in self._auxiliary_observations.items():

            print('Sampling: %s' %k)

            # set the luminosities and distances to
            # auxilary sampler just in case
            # they are needed
            
            v.set_luminosity(luminosities)
            v.set_distance(distances)

            # sample the true and obs
            # values which are held internally
            v.true_sampler(size=n)
            v.observation_sampler(size=n)

            # check to make sure we sampled!
            assert v.true_values is not None and len(v.true_values) == n
            assert v.obs_values is not None and len(v.obs_values) == n

            # append these values to a dict
            auxiliary_quantities[k] = {'true_values': v.true_values,
                                       'obs_values': v.obs_values,
                                       'sigma': v.sigma }
            
            

        # now draw all the observed fluxes
        # this is homoskedastic for now
        log10_fluxes_obs = self.draw_log10_fobs(fluxes, flux_sigma, size=n)

        # compute the detection probability  for the observed values
        detection_probability = self._prob_det(log10_fluxes_obs, np.log10(boundary), strength)

        # now select them


        if not hard_cut:

            selection = []
            for p in detection_probability:

                # make a bernoulli draw given the detection probability
                if stats.bernoulli.rvs(p) == 1:

                    selection.append(True)

                else:

                    selection.append(False)

            selection = np.array(selection)

        else:

            selection = np.power(10, log10_fluxes_obs) >= boundary 
            

 
 
        if sum(selection) == n:
            print('NO HIDDEN OBJECTS')


        if distance_probability is not None:

            known_distances = []
            known_distance_idx = []
            unknown_distance_idx = []
            
            assert (distance_probability>=0) and (distance_probability <=1.), 'the distance detection must be between 0 and 1'

            for i, d in enumerate(distances[selection]):

                # see if we detect the distance
                if stats.bernoulli.rvs(distance_probability) == 1:

                    known_distances.append(d)
                    known_distance_idx.append(i)

                else:

                    unknown_distance_idx.append(i)

            print('Detected %d distances' %len(known_distances))

        else:


            known_distances = distances[selection]
            known_distance_idx = [i for i in range(sum(selection))]
            unknown_distance_idx = []

        known_distances = np.array(known_distances)
        known_distance_idx = np.array(known_distance_idx)
        unknown_distance_idx = np.array(unknown_distance_idx)
                    
        try:
            print('Deteced %d objects or to a distance of %.2f' %(sum(selection), max(known_distances)))

        except:
            print('No Objects detected')
        # return a Population object

        return Population(
            luminosities=luminosities,
            distances=distances,
            known_distances=known_distances,
            known_distance_idx=known_distance_idx,
            unknown_distance_idx=unknown_distance_idx,
            fluxes=fluxes,
            flux_obs=np.power(10, log10_fluxes_obs),
            selection=selection,
            flux_sigma=flux_sigma,
            r_max = self._r_max,
            n_model=self._n_model,
            lf_params=self._lf_params,
            spatial_params=self._spatial_params,
            model_spaces=self._model_spaces,
            boundary=boundary,
            strength=strength,
            seed=self._seed,
            name=self._name,
            spatial_form=self._spatial_form,
            lf_form=self._lf_form,
            auxiliary_quantities=auxiliary_quantities
        )


    
    def display(self):
        """
        Display the simulation parameters
        
        """

        out={'parameter':[], 'value':[]}

        display(Markdown('## Luminosity Function'))
        for k,v in self._lf_params.items():

             out['parameter'].append(k)
             out['value'].append(v)

        display(Math(self._lf_form))
        display(pd.DataFrame(out))
        out={'parameter':[], 'value':[]}

        display(Markdown('## Spatial Function'))


        for k,v in self._spatial_params.items():

            out['parameter'].append(k)
            out['value'].append(v)

        display(Math(self._spatial_form))
        display(pd.DataFrame(out))


    def generate_stan_code(self, stan_gen, **kwargs):

        pass



import numpy as np
import multiprocessing as mp


def sample_one(ymax, idx1,  xbreak,  idx2,  xmin,  xmax, i):
    
    np.random.seed(int((i+1)*1000))
    
    flag = True
    while(flag):
        y_guess = np.random.uniform(0, ymax)
        x_guess =  np.random.uniform(1, xmax/xmin)

        x_test = bpl( x_guess,  idx1,  xbreak/xmin,  idx2 )
        if y_guess <= x_test:
            flag = False
    return x_guess

def bpl( x,  idx1,  xbreak,  idx2 ):
    if x < xbreak:
        return pow(x/xbreak,-idx1)
    else:
        return pow(x/xbreak,-idx2)





from popsynth.population_synth import PopulationSynth


class BPLPopulation(PopulationSynth):

    def __init__(self, Lmin, alpha, Lbreak, beta, Lmax,  r_max=10, seed=1234, name='_pareto'):

        PopulationSynth.__init__(self, r_max, seed, name)

        self.set_luminosity_function_parameters(Lmin=Lmin, alpha=alpha, Lbreak=Lbreak, beta=beta, Lmax=Lmax)


        self._lf_form = r"\frac{\alpha L_{\rm min}^{\alpha}}{L^{\alpha+1}}"
    

    def phi(self, L):

        out = np.zeros_like(L)

        idx = L>=self.Lmin
        
        out[idx] =  self.alpha*self.Lmin**self.alpha / L[idx]**(self.alpha+1)

        return out
        
    def draw_luminosity(self, size=1):

        
        ymax = bpl(1., self.alpha,  self.Lbreak/self.Lmin,  self.beta )
    
        result_list = []
    
    
        def log_result(result):
            # This is called whenever foo_pool(i) returns a result.
            # result_list is modified only by the main process, not the pool workers.
            result_list.append(result)
                
        with mp.Pool(8) as pool:
    
            for i in range(size):
                pool.apply_async(sample_one, args=(ymax, self.alpha,  self.Lbreak,  self.beta,  self.Lmin,  self.Lmax, i), callback = log_result)
            pool.close()
            pool.join()
  

    
        return np.array(result_list) * self.Lmin
    
        
    def __get_Lmin(self):
             """Calculates the 'Lmin' property."""
             return self._lf_params['Lmin']

    def ___get_Lmin(self):
         """Indirect accessor for 'Lmin' property."""
         return self.__get_Lmin()

    def __set_Lmin(self, Lmin):
         """Sets the 'Lmin' property."""
         self.set_luminosity_function_parameters(alpha=self.alpha, Lmin=Lmin)

    def ___set_Lmin(self, Lmin):
         """Indirect setter for 'Lmin' property."""
         self.__set_Lmin(Lmin)

    Lmin = property(___get_Lmin, ___set_Lmin,
                     doc="""Gets or sets the Lmin.""")


    def __get_Lmax(self):
             """Calculates the 'Lmax' property."""
             return self._lf_params['Lmax']

    def ___get_Lmax(self):
         """Indirect accessor for 'Lmax' property."""
         return self.__get_Lmax()

    def __set_Lmax(self, Lmax):
         """Sets the 'Lmax' property."""
         self.set_luminosity_function_parameters(alpha=self.alpha, Lmax=Lmax)

    def ___set_Lmax(self, Lmax):
         """Indirect setter for 'Lmax' property."""
         self.__set_Lmax(Lmax)

    Lmax = property(___get_Lmax, ___set_Lmax,
                     doc="""Gets or sets the Lmax.""")



    

    def __get_Lbreak(self):
        """Calculates the 'Lbreak' property."""
        return self._lf_params['Lbreak']

    def ___get_Lbreak(self):
         """Indirect accessor for 'Lbreak' property."""
         return self.__get_Lbreak()

    def __set_Lbreak(self, Lbreak):
         """Sets the 'Lbreak' property."""
         self.set_luminosity_function_parameters(alpha=self.alpha, Lbreak=Lbreak)

    def ___set_Lbreak(self, Lbreak):
         """Indirect setter for 'Lbreak' property."""
         self.__set_Lbreak(Lbreak)

    Lbreak = property(___get_Lbreak, ___set_Lbreak,
                     doc="""Gets or sets the Lbreak.""")


    
    def __get_alpha(self):
             """Calculates the 'alpha' property."""
             return self._lf_params['alpha']

    def ___get_alpha(self):
         """Indirect accessor for 'alpha' property."""
         return self.__get_alpha()

    def __set_alpha(self, alpha):
         """Sets the 'alpha' property."""
         self.set_luminosity_function_parameters(alpha=alpha,Lmin=self.Lmin)

    def ___set_alpha(self, alpha):
         """Indirect setter for 'alpha' property."""
         self.__set_alpha(alpha)

    alpha = property(___get_alpha, ___set_alpha,
                     doc="""Gets or sets the alpha.""")


    def __get_beta(self):
             """Calculates the 'beta' property."""
             return self._lf_params['beta']

    def ___get_beta(self):
         """Indirect accessor for 'beta' property."""
         return self.__get_beta()

    def __set_beta(self, beta):
         """Sets the 'beta' property."""
         self.set_luminosity_function_parameters(beta=beta,Lmin=self.Lmin)

    def ___set_beta(self, beta):
         """Indirect setter for 'beta' property."""
         self.__set_beta(beta)

    beta = property(___get_beta, ___set_beta,
                     doc="""Gets or sets the beta.""")






import scipy.special as sf
import numpy as np

from popsynth.population_synth import PopulationSynth

class SchechterPopulation(PopulationSynth):

    def __init__(self, Lmin, alpha, r_max=10, seed=1234, name='_shechter'):

        PopulationSynth.__init__(self, r_max, seed, name)

        self.set_luminosity_function_parameters(Lmin=Lmin, alpha=alpha)

        self._lf_form = r'\frac{1}{L_{\rm min}^{1+\alpha} \Gamma\left(1+\alpha\right)} L^{\alpha} \exp\left[ - \frac{L}{L_{\rm min}}\right]'
    
    def phi(self, L):
        return  L**self.alpha * np.exp(-L/self.Lmin)/(self.Lmin**(1+self.alpha) * sf.gamma(1+self.alpha))

    def draw_luminosity(self, size=1):

        xs = (1 - np.random.uniform(size=size))
        return self.Lmin*sf.gammaincinv(1+self.alpha, xs)
        
    def __get_Lmin(self):
             """Calculates the 'Lmin' property."""
             return self._lf_params['Lmin']

    def ___get_Lmin(self):
         """Indirect accessor for 'Lmin' property."""
         return self.__get_Lmin()

    def __set_Lmin(self, Lmin):
         """Sets the 'Lmin' property."""
         self.set_luminosity_function_parameters(alpha=self.alpha, Lmin=Lmin)

    def ___set_Lmin(self, Lmin):
         """Indirect setter for 'Lmin' property."""
         self.__set_Lmin(Lmin)

    Lmin = property(___get_Lmin, ___set_Lmin,
                     doc="""Gets or sets the Lmin.""")


    
    def __get_alpha(self):
             """Calculates the 'alpha' property."""
             return self._lf_params['alpha']

    def ___get_alpha(self):
         """Indirect accessor for 'alpha' property."""
         return self.__get_alpha()

    def __set_alpha(self, alpha):
         """Sets the 'alpha' property."""
         self.set_luminosity_function_parameters(alpha=alpha,Lmin=self.Lmin)

    def ___set_alpha(self, alpha):
         """Indirect setter for 'alpha' property."""
         self.__set_alpha(alpha)

    alpha = property(___get_alpha, ___set_alpha,
                     doc="""Gets or sets the alpha.""")

